from collections import OrderedDict

###################################################################################
#  词法分析器                                                                     #
###################################################################################
INTEGER = 'INTEGER'
REAL = 'REAL'
INTEGER_CONST = 'INTEGER_CONST'
REAL_CONST = 'REAL_CONST'
PLUS = 'PLUS'
MINUS = 'MINUS'
MUL = 'MUL'
INTEGER_DIV = 'INTEGER_DIV'
FLOAT_DIV = 'FLOAT_DIV'
LPAREN = 'LPAREN'
RPAREN = 'RPAREN'
ID = 'ID'
ASSIGN = 'ASSIGN'
BEGIN = 'BEGIN'
END = 'END'
SEMI = 'SEMI'
DOT = 'DOT'
PROGRAM = 'PROGRAM'
VAR = 'VAR'
PROCEDURE = 'PROCEDURE'  # 过程
COLON = 'COLON'
COMMA = 'COMMA'
EOF = 'EOF'


class Token:
    def __init__(self, value_type, value):
        self.value_type = value_type
        self.value = value

    def __str__(self):
        return 'Token({value_type},{value})'.format(value_type=self.value_type, value=self.value)

    def __repr__(self):
        return self.__str__()


RESERVED_KEYWORDS = {
    'PROGRAM': Token('PROGRAM', 'PROGRAM'),
    'PROCEDURE': Token('PROCEDURE', 'PROCEDURE'),  # 保留字
    'VAR': Token('VAR', 'VAR'),
    'DIV': Token('INTEGER_DIV', 'DIV'),
    'INTEGER': Token('INTEGER', 'INTEGER'),
    'REAL': Token('REAL', 'REAL'),
    'BEGIN': Token('BEGIN', 'BEGIN'),
    'END': Token('END', 'END'),
}


class Lexer:
    def __init__(self, text):
        self.text = text
        self.position = 0
        self.current_char = self.text[self.position]

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def number(self):
        result = ''
        while self.current_char is not None and self.current_char.isdigit():
            result += self.current_char
            self.advance()
            if self.current_char == '.':
                result += self.current_char
                self.advance()
                while self.current_char is not None and self.current_char.isdigit():
                    result += self.current_char
                    self.advance()
                return Token(REAL_CONST, float(result))
        return Token(INTEGER_CONST, int(result))

    def skip_comment(self):
        while self.current_char != '}':
            self.advance()
        self.advance()

    def _id(self):
        result = ''
        if self.current_char == '_':
            result = '_'
            self.advance()
        while self.current_char is not None and self.current_char.isalnum():
            result += self.current_char
            self.advance()
        token = RESERVED_KEYWORDS.get(result.upper(), Token('ID', result.lower()))
        return token

    def peek(self):
        pos = self.position + 1
        if pos >= len(self.text):
            return None
        else:
            return self.text[pos]

    def skip_whitespace(self):
        while self.current_char is not None and self.current_char.isspace():
            self.advance()

    def advance(self):
        self.position += 1
        if self.position >= len(self.text):
            self.current_char = None
        else:
            self.current_char = self.text[self.position]

    def get_next_token(self):
        while self.current_char is not None:
            if self.current_char == '{':
                self.advance()
                self.skip_comment()
                continue
            if self.current_char.isspace():
                self.skip_whitespace()
                continue
            if self.current_char.isalpha() or self.current_char == '_':
                return self._id()
            if self.current_char == ':' and self.peek() == '=':
                self.advance()
                self.advance()
                return Token(ASSIGN, ':=')
            if self.current_char == ';':
                self.advance()
                return Token(SEMI, ';')
            if self.current_char == '.':
                self.advance()
                return Token(DOT, '.')
            if self.current_char.isdigit():
                return self.number()
            if self.current_char == ':':
                self.advance()
                return Token(COLON, ':')
            if self.current_char == ',':
                self.advance()
                return Token(COMMA, ',')
            if self.current_char == '+':
                self.advance()
                return Token(PLUS, '+')
            if self.current_char == '-':
                self.advance()
                return Token(MINUS, '-')
            if self.current_char == '*':
                self.advance()
                return Token(MUL, '*')
            if self.current_char == '/':
                self.advance()
                return Token(FLOAT_DIV, '/')
            if self.current_char == '(':
                self.advance()
                return Token(LPAREN, '(')
            if self.current_char == ')':
                self.advance()
                return Token(RPAREN, ')')
            self.error()
        return Token(EOF, None)


###################################################################################
#  语法分析器                                                                     #
###################################################################################
class AST:
    pass


class Program(AST):
    def __init__(self, name, block):
        self.name = name
        self.block = block


class ProcedureDecl(AST):  # 添加过程声明节点
    def __init__(self, name, block_node):
        self.name = name  # 名称
        self.block_node = block_node  # 块节点


class Block(AST):
    def __init__(self, declarations, compound_statement):
        self.declarations = declarations
        self.compound_statement = compound_statement


class VarDecl(AST):
    def __init__(self, var_node, type_node):
        self.var_node = var_node
        self.type_node = type_node


class Type(AST):
    def __init__(self, token):
        self.token = token
        self.name = token.value


class Compound(AST):
    def __init__(self):
        self.children = []


class Assign(AST):
    def __init__(self, left, operator, right):
        self.left = left
        self.token = self.operator = operator
        self.right = right


class Variable(AST):
    def __init__(self, token):
        self.token = token
        self.name = token.value


class NoOperator(AST):
    pass


class BinaryOperator(AST):
    def __init__(self, left, operator, right):
        self.left = left
        self.token = self.operator = operator
        self.right = right


class Num(AST):
    def __init__(self, token):
        self.token = token
        self.value = token.value


class UnaryOperator(AST):
    def __init__(self, operator, expr):
        self.token = self.operator = operator
        self.expr = expr


class Parser:
    def __init__(self, lexer):
        self.lexer = lexer
        self.current_token = lexer.get_next_token()

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def eat(self, value_type):
        if self.current_token.value_type == value_type:
            self.current_token = self.lexer.get_next_token()
        else:
            self.error()

    def program(self):
        self.eat(PROGRAM)
        var_node = self.variable()
        program_name = var_node.name
        self.eat(SEMI)
        block_node = self.block()
        node = Program(program_name, block_node)
        self.eat(DOT)
        return node

    def block(self):
        declarations = self.declarations()
        compound_statement = self.compound_statement()
        node = Block(declarations, compound_statement)
        return node

    def declarations(self):  # 修改创建声明节点的方法
        declarations = []
        if self.current_token.value_type == VAR:
            self.eat(VAR)
            while self.current_token.value_type == ID:
                declarations.extend(self.variable_declaration())
                self.eat(SEMI)
        while self.current_token.value_type == PROCEDURE:  # 当前记号类型是过程时
            self.eat(PROCEDURE)  # 验证过程类型
            procedure_name = self.current_token.value  # 获取过程名称
            self.eat(ID)  # 验证过程名称
            self.eat(SEMI)  # 验证分号
            block_node = self.block()  # 获取过程中的块
            procedure_decl = ProcedureDecl(procedure_name, block_node)  # 由过程名称和块组成过程声明对象
            declarations.append(procedure_decl)  # 声明列表末尾添加新的过程声明
            self.eat(SEMI)  # 验证分号
        return declarations

    def variable_declaration(self):
        var_nodes = [Variable(self.current_token)]
        self.eat(ID)
        while self.current_token.value_type == COMMA:
            self.eat(COMMA)
            var_nodes.append(Variable(self.current_token))
            self.eat(ID)
        self.eat(COLON)
        type_node = self.type_spec()
        var_declarations = [VarDecl(var_node, type_node) for var_node in var_nodes]
        return var_declarations

    def type_spec(self):
        token = self.current_token
        if token.value_type == INTEGER:
            self.eat(INTEGER)
        else:
            self.eat(REAL)
        node = Type(token)
        return node

    def compound_statement(self):
        self.eat(BEGIN)
        nodes = self.statement_list()
        self.eat(END)

        root = Compound()
        root.children = nodes
        return root

    def statement_list(self):
        node = self.statement()
        nodes = [node]
        while self.current_token.value_type == SEMI:
            self.eat(SEMI)
            nodes.append(self.statement())
        if self.current_token.value_type == ID:
            self.error()
        return nodes

    def statement(self):
        if self.current_token.value_type == BEGIN:
            node = self.compound_statement()
        elif self.current_token.value_type == ID:
            node = self.assignment_statement()
        else:
            node = self.empty()
        return node

    def assignment_statement(self):
        left = self.variable()
        token = self.current_token
        self.eat(ASSIGN)
        right = self.expr()
        node = Assign(left, token, right)
        return node

    def empty(self):
        return NoOperator()

    def variable(self):
        node = Variable(self.current_token)
        self.eat(ID)
        return node

    def factor(self):
        current_token = self.current_token

        if current_token.value_type == PLUS:
            self.eat(PLUS)
            node = UnaryOperator(current_token, self.factor())
            return node
        elif current_token.value_type == MINUS:
            self.eat(MINUS)
            node = UnaryOperator(current_token, self.factor())
            return node
        elif current_token.value_type == INTEGER_CONST:
            self.eat(INTEGER_CONST)
            return Num(current_token)
        elif current_token.value_type == REAL_CONST:
            self.eat(REAL_CONST)
            return Num(current_token)
        elif current_token.value_type == LPAREN:
            self.eat(LPAREN)
            node = self.expr()
            self.eat(RPAREN)
            return node
        else:
            node = self.variable()
            return node

    def term(self):
        node = self.factor()
        while self.current_token.value_type in (MUL, INTEGER_DIV, FLOAT_DIV):
            token = self.current_token
            if token.value_type == MUL:
                self.eat(MUL)
            elif token.value_type == INTEGER_DIV:
                self.eat(INTEGER_DIV)
            elif token.value_type == FLOAT_DIV:
                self.eat(FLOAT_DIV)
            node = BinaryOperator(node, token, self.factor())
        return node

    def expr(self):
        node = self.term()
        while self.current_token.value_type in (PLUS, MINUS):
            token = self.current_token
            if token.value_type == PLUS:
                self.eat(PLUS)
            elif token.value_type == MINUS:
                self.eat(MINUS)
            node = BinaryOperator(node, token, self.term())
        return node

    def parser(self):
        node = self.program()
        if self.current_token.value_type != EOF:
            self.error()
        return node


###################################################################################
#  AST访问器                                                                      #
###################################################################################
class NodeVisitor:
    def visit(self, node):
        method_name = 'visit_' + type(node).__name__
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        raise Exception(f'未找到“visit_{type(node).__name__}()”方法！')


###################################################################################
#  符号和符号表                                                                   #
###################################################################################

class Symbol:
    def __init__(self, name, symbol_type=None):
        self.name = name
        self.symbol_type = symbol_type


class BuiltinTypeSymbol(Symbol):
    def __init__(self, name):
        super().__init__(name)

    def __str__(self):
        return self.name

    __repr__ = __str__


class VarSymbol(Symbol):
    def __init__(self, name, symbol_type):
        super().__init__(name, symbol_type)

    def __str__(self):
        return f'<{self.name}:{self.symbol_type}>'

    __repr__ = __str__


class SymbolTable:
    def __init__(self):
        self._symbols = OrderedDict()
        self.__init_builtins()

    def __init_builtins(self):
        self.define(BuiltinTypeSymbol('INTEGER'))
        self.define(BuiltinTypeSymbol('REAL'))

    def __str__(self):
        return f'Symbols:{[value for value in self._symbols.values()]}'

    __repr__ = __str__

    def define(self, symbol):
        print(f'定义：{symbol}')
        self._symbols[symbol.name] = symbol

    def lookup(self, name):
        print(f'查询：{name}')
        return self._symbols.get(name)


class SymbolTableBuilder(NodeVisitor):
    def __init__(self):
        self.symbol_table = SymbolTable()

    def visit_Program(self, node):
        self.visit(node.block)

    def visit_Block(self, node):
        for declaration in node.declarations:
            self.visit(declaration)
        self.visit(node.compound_statement)

    def visit_ProcedureDecl(self, node):  # 添加访问过程声明的方法
        pass  # 暂不处理

    def visit_VarDecl(self, node):
        type_name = node.type_node.name
        symbol_type = self.symbol_table.lookup(type_name)
        var_name = node.var_node.name
        var_symbol = VarSymbol(var_name, symbol_type)
        self.symbol_table.define(var_symbol)

    def visit_Compound(self, node):
        for child in node.children:
            self.visit(child)

    def visit_NoOperator(self, node):
        pass

    def visit_BinaryOperator(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit_Num(self, node):
        pass

    def visit_UnaryOperator(self, node):
        self.visit(node.expr)

    def visit_Assign(self, node):
        var_name = node.left.name
        var_symbol = self.symbol_table.lookup(var_name)
        if var_symbol is None:
            raise NameError(f'错误的标识符：{repr(var_name)}')
        self.visit(node.right)

    def visit_Variable(self, node):
        var_name = node.name
        var_symbol = self.symbol_table.lookup(var_name)
        if var_symbol is None:
            raise NameError(f'错误的标识符：{repr(var_name)}')


###################################################################################
#  解释器                                                                         #
###################################################################################
class Interpreter(NodeVisitor):
    def __init__(self, tree):
        self.tree = tree
        self.GLOBAL_MEMORY = OrderedDict()

    def visit_Program(self, node):
        self.visit(node.block)

    def visit_Block(self, node):
        for declaration in node.declarations:
            self.visit(declaration)
        self.visit(node.compound_statement)

    def visit_ProcedureDecl(self, node):  # 添加访问过程声明的方法
        pass  # 暂不处理

    def visit_VarDecl(self, node):
        pass

    def visit_Type(self, node):
        pass

    def visit_Compound(self, node):
        for child in node.children:
            self.visit(child)

    def visit_Assign(self, node):
        var_name = node.left.name
        self.GLOBAL_MEMORY[var_name] = self.visit(node.right)

    def visit_Variable(self, node):
        var_name = node.name
        value = self.GLOBAL_MEMORY.get(var_name)
        if value is None:
            raise NameError(f'错误的标识符：{repr(var_name)}')
        else:
            return value

    def visit_NoOperator(self, node):
        pass

    def visit_BinaryOperator(self, node):
        if node.operator.value_type == PLUS:
            return self.visit(node.left) + self.visit(node.right)
        elif node.operator.value_type == MINUS:
            return self.visit(node.left) - self.visit(node.right)
        elif node.operator.value_type == MUL:
            return self.visit(node.left) * self.visit(node.right)
        elif node.operator.value_type == INTEGER_DIV:
            return self.visit(node.left) // self.visit(node.right)
        elif node.operator.value_type == FLOAT_DIV:
            return self.visit(node.left) / self.visit(node.right)

    def visit_Num(self, node):
        return node.value

    def visit_UnaryOperator(self, node):
        if node.operator.value_type == PLUS:
            return +self.visit(node.expr)
        if node.operator.value_type == MINUS:
            return -self.visit(node.expr)

    def interpret(self):
        tree = self.tree
        if tree is None:
            return ''
        return self.visit(tree)


def main():
    import sys
    text = open(sys.argv[1], 'r').read()  # 打开系统命令中的第二个参数（Pascal文件）并读取内容
    lexer = Lexer(text)
    parser = Parser(lexer)
    tree = parser.parser()
    symtab_builder = SymbolTableBuilder()
    symtab_builder.visit(tree)
    print('符号表中的内容：\n', symtab_builder.symbol_table)
    print('-----------------------------------------------')
    print('开始解释代码...')
    interpreter = Interpreter(tree)
    interpreter.interpret()
    print('全局存储中的内容：')
    for k, v in interpreter.GLOBAL_MEMORY.items():
        print(f'{k}：{v}')


if __name__ == '__main__':
    main()
